/***************************************************************************
 *   Copyright (C) 2006 by Borko Boskovic                                  *
 *   borko.boskovic@uni-mb.si                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <glib/gprintf.h>

#include "board.h"
#include "asm.h"

#define shiftDown(b)      ((b)>>8)		/*< ??? */
#define shift2Down(b)     ((b)>>16)		/*< ??? */
#define shiftUp(b)        ((b)<<8)		/*< ??? */
#define shift2Up(b)       ((b)<<16)		/*< ??? */
#define shiftRight(b)     (((b)<<1)&NOTFILEA) 	/*< ??? */
#define shift2Right(b)    (((b)<<2)&NOTFILEAB)	/*< ??? */
#define shiftLeft(b)      (((b)>>1)&NOTFILEH)	/*< ??? */
#define shift2Left(b)     (((b)>>2)&NOTFILEGH)	/*< ??? */
#define shiftUpLeft(b)    (((b)<<7)&NOTFILEH)	/*< ??? */
#define shiftUpRight(b)   (((b)<<9)&NOTFILEA)	/*< ??? */
#define shiftDownLeft(b)  (((b)>>9)&NOTFILEH)	/*< ??? */
#define shiftDownRight(b) (((b)>>7)&NOTFILEA)	/*< ??? */

#define uChangePlayer(board) {board->player ^= 1;board->opponent ^= 1;} 	/*< ??? */
#define changePlayer(board) {uChangePlayer(board);board->key ^= playerKey;}	/*< ??? */

#define WHITE_CASTLING 3		/*< ??? */
#define WHITE_CASTLING_KING_SIDE 1	/*< ??? */
#define WHITE_CASTLING_QUEEN_SIDE 2	/*< ???	*/
#define BLACK_CASTLING 12		/*< ???	*/
#define BLACK_CASTLING_KING_SIDE 4	/*< ??? */
#define BLACK_CASTLING_QUEEN_SIDE 8	/*< ??? */

gint64 squareKey[14][64];		/*< ??? */
gint64 playerKey;			/*< ??? */

guint64 attackPawn[2][64];		/*< ??? */
guint64 attackKnight[64];		/*< ??? */
guint64 attackBishopL45[64][64];	/*< ??? */
guint64 attackBishopR45[64][64];	/*< ??? */
guint64 attackRookL00[64][64];		/*< ??? */
guint64 attackRookL90[64][64];		/*< ??? */
guint64 attackKing[64];			/*< ??? */

guint64 maskKnight[64];			/*< ??? */
guint64 maskBishop[64];			/*< ??? */
guint64 maskRook[64];			/*< ??? */
guint64 maskQueen[64];			/*< ??? */
guint64 maskKingSafe[64];

gint mobilityBishopL45[64][64];
gint mobilityBishopR45[64][64];
gint mobilityRookL00[64][64];
gint mobilityRookL90[64][64];
gint mobilityKnight8[64];

gboolean isInit = FALSE;		/*< ??? */

const guint rotateL90[64] = {		/*< ??? */
	h1,h2,h3,h4,h5,h6,h7,h8,
	g1,g2,g3,g4,g5,g6,g7,g8,
	f1,f2,f3,f4,f5,f6,f7,f8,
	e1,e2,e3,e4,e5,e6,e7,e8,
	d1,d2,d3,d4,d5,d6,d7,d8,
	c1,c2,c3,c4,c5,c6,c7,c8,
	b1,b2,b3,b4,b5,b6,b7,b8,
	a1,a2,a3,a4,a5,a6,a7,a8,
};

const guint rotateR45[64] = {		/*< ??? */
	a1,b8,c7,d6,e5,f4,g3,h2,
	a2,b1,c8,d7,e6,f5,g4,h3,
	a3,b2,c1,d8,e7,f6,g5,h4,
	a4,b3,c2,d1,e8,f7,g6,h5,
	a5,b4,c3,d2,e1,f8,g7,h6,
	a6,b5,c4,d3,e2,f1,g8,h7,
	a7,b6,c5,d4,e3,f2,g1,h8,
	a8,b7,c6,d5,e4,f3,g2,h1
};

const guint rotateL45[64] = {		/*< ??? */
	a2,b3,c4,d5,e6,f7,g8,h1,
	a3,b4,c5,d6,e7,f8,g1,h2,
	a4,b5,c6,d7,e8,f1,g2,h3,
	a5,b6,c7,d8,e1,f2,g3,h4,
	a6,b7,c8,d1,e2,f3,g4,h5,
	a7,b8,c1,d2,e3,f4,g5,h6,
	a8,b1,c2,d3,e4,f5,g6,h7,
	a1,b2,c3,d4,e5,f6,g7,h8
};

const guint shiftR45[64] = {		/*< ??? */
	1,58,51,44,37,30,23,16,
	9, 1,58,51,44,37,30,23,
	17, 9, 1,58,51,44,37,30,
	25,17, 9, 1,58,51,44,37,
	33,25,17, 9, 1,58,51,44,
	41,33,25,17, 9, 1,58,51,
	49,41,33,25,17, 9, 1,58,
	57,49,41,33,25,17, 9, 1
};

const guint shiftL45[64] = {   		/*< ??? */
	9,17,25,33,41,49,57, 1,
	17,25,33,41,49,57, 1,10,
	25,33,41,49,57, 1,10,19,
	33,41,49,57, 1,10,19,28,
	41,49,57, 1,10,19,28,37,
	49,57, 1,10,19,28,37,46,
	57, 1,10,19,28,37,46,55,
	1,10,19,28,37,46,55,64
};

/**
 * inline guint64 attackR45(guint square, gint64 piecesR45)
 * ???
 * \param square ???
 * \param piecesR45 ???
 * \return ???
 */
inline guint64 attackR45(guint square, gint64 piecesR45){
	return attackBishopR45[square][((piecesR45)>>shiftR45[square])&63];
}

/**
 * inline guint64 attackL45(guint square, guint64 piecesL45)
 * ???
 * \param square ???
 * \param piecesL45 ???
 * \return ???
 */
inline guint64 attackL45(guint square, guint64 piecesL45){
	return attackBishopL45[square][((piecesL45)>>shiftL45[square])&63];
}

/**
 * inline guint64 attackRank(guint square, guint64 pieces){
 * \param square ???
 * \param pieces ???
 * \return ???
 */
inline guint64 attackRank(guint square, guint64 pieces){
	return attackRookL00[(square)&7][((pieces)>>(((square)&~7)+1))&63]<<((square)&~7);
}

/**
 * inline guint64 attackFile(guint square, guint64 piecesL90)
 * \param square
 * \param piecesL90
 * \return
 */
inline guint64 attackFile(guint square, guint64 piecesL90){
	return attackRookL90[square][((piecesL90)>>((((square)&7)<<3)+1))&63];
}

/**
 * inline guint64 attackRook(Board* board, guint square)
 * ???
 * \param board ???
 * \param square ???
 * \return ???
 */
inline guint64 attackRook(const Board* board, guint square){
	return attackRank(square,board->notFreeSquares)|attackFile(square,board->piecesRL90);
}

/**
 * inline guint64 attackBishop(Board* board, guint square)
 * ???
 * \param board ???
 * \param square ???
 * \return ???
 */
inline guint64 attackBishop(const Board* board, guint square){
	return attackR45(square,board->piecesRR45)|attackL45(square,board->piecesRL45);
}

/**
 * inline guint64 attackQueen(Board* board, guint square)
 * ???
 * \param board ???
 * \param square ???
 * \return ???
 */
inline guint64 attackQueen(const Board* board, guint square){
	return attackRook(board, square) | attackBishop(board, square);
}

inline gint mobilityR45(gint square, guint64 piecesR45){
	return mobilityBishopR45[square][((piecesR45)>>shiftR45[square])&63];
}

inline gint mobilityL45(gint square, guint64 piecesL45){
	return mobilityBishopL45[square][((piecesL45)>>shiftL45[square])&63];
}

inline gint mobilityRank(gint square, guint64 pieces){
	return mobilityRookL00[(square)&7][((pieces)>>(((square)&~7)+1))&63];
}

inline gint mobilityFile(gint square, guint64 piecesL90){
	return mobilityRookL90[square][((piecesL90)>>((((square)&7)<<3)+1))&63];
}

inline gint mobilityBishop(const Board* board, gint square){ 
	return mobilityR45(square,board->piecesRR45)+\
			mobilityL45(square,board->piecesRL45);
}

inline gint mobilityRook(const Board* board, gint square){
	return mobilityRank(square,board->notFreeSquares)+\
			mobilityFile(square,board->piecesRL90);
}

inline gint mobilityQueen(const Board* board, gint square){
	return mobilityRook(board, square) + mobilityBishop(board, square);
}

inline gint mobilityKnight(gint square){
	return mobilityKnight8[square];
}

inline gint numberOfBits64(guint64 b){
    gint n;
    for(n = 0; b != 0; n++, b = b & (b - 1));
    return n;
}

void boardBitPrint(guint64 x){
	gint i, index;
	gchar stype[64];
	for(i=0; i<64; i++) stype[i] = '0';
	stype[64] = '\0';
	while(x != 0){
		index = getFirstBitAndClear64(&x);
		stype[63-index] = '1';
	}
	g_printf("%s\n\n",stype);
}

void boardChessBitPrint(guint64 x){
	gint i, index;
	gchar stype[72];
	for(i=0; i<71; i++) stype[i] = '0';
	for(i=8; i<71; i+=9) stype[i] = '\n';
	stype[71] = '\0';
	while(x != 0){
		index = getFirstBitAndClear64(&x);
		stype[70-rank(index)*9-(7-file(index))] = '1';
	}
	g_printf("%s\n\n",stype);
}

Board* boardNew(){
	g_assert(isInit);
	Board * board;
	board = malloc(sizeof(Board));
	board->pawns = &board->pieces[PAWN];
	board->knights = &board->pieces[KNIGHT];
	board->bishops = &board->pieces[BISHOP];
	board->rooks = &board->pieces[ROOK];
	board->queens = &board->pieces[QUEEN];
	board->king = &board->pieces[KING];
	boardStartPosition(board);
	return board;
}

void boardFree(Board* board){
	free(board);
}

/**
 * inline void boardEmptyPosition(Board* board)
 * ???
 * \param board ???
 */
inline void boardEmptyPosition(Board* board){
	gint i;
	for(i=WHITE; i<=BLACK; i++){
		board->pawns[i] = board->knights[i] = board->bishops[i] = 0;
		board->rooks[i] = board->queens[i] = board->king[i] = 0;
		board->pieces[i] = 0;
		board->material[i] = board->numberOfPieces[i] = 0;
	}
	board->key = board->piecesRL90 = board->piecesRL45 = board->piecesRR45 = 0;
	board->ply = board->player = board->enPassant = board->castling = 0;
	board->check = 0;
	board->opponent = 1;
	board->notFreeSquares = 0;
	board->freeSquares = ~0;
	for(i=0; i<MAX_PLY; i++){
		board->historyOfKeys[i] = 0;
		board->numberOfStaticMoves[i] = 0;
		board->undoCapturePiece[i] = board->undoCheck[i] = 0;
		board->undoCastling[i] = board->undoEnPassant[i] = 0;
	}
	for(i=0; i<64; i++) board->types[i] = EMPTY;
}

/**
 * inline void boardAddPiece(Board* board, guint piece, guint square)
 * ???
 * \param board ???
 * \param piece ???
 * \param square ???
 */
inline void boardAddPiece(Board* board, guint piece, guint square){
	guint64 bbsquare;
	guint player;
	bbsquare = 1ull << square;
	player = piece%2;
	board->material[player] += pieceValue[piece];
	board->numberOfPieces[player] ++;
	board->pieces[player] ^= bbsquare;
	board->pieces[piece] ^= bbsquare;
	board->notFreeSquares ^= bbsquare;
	board->freeSquares = ~board->notFreeSquares;
	board->piecesRL90 ^= 1ull << rotateL90[square];
	board->piecesRL45 ^= 1ull << rotateL45[square];
	board->piecesRR45 ^= 1ull << rotateR45[square];
	board->key ^= squareKey[piece][square];
	board->types[square] = piece;
}

/**
 * inline guint boardRemovePiece(Board* board, guint square)
 * ???
 * \param board ???
 * \param square ???
 */
inline guint boardRemovePiece(Board* board, guint square){
	guint64 bbsquare;
	guint player;
	guint piece = board->types[square];
	g_assert(piece != EMPTY);
	bbsquare = 1ull << square;
	player = piece % 2;
	board->material[player] -= pieceValue[piece];
	board->numberOfPieces[player] --;
	board->pieces[player] ^= bbsquare;
	board->pieces[piece] ^= bbsquare;
	board->notFreeSquares ^= bbsquare;
	board->freeSquares = ~board->notFreeSquares;
	board->piecesRL90 ^= 1ull << rotateL90[square];
	board->piecesRL45 ^= 1ull << rotateL45[square];
	board->piecesRR45 ^= 1ull << rotateR45[square];
	board->key ^= squareKey[piece][square];
	board->types[square] = EMPTY;
	return piece;
}

void boardStartPosition(Board* board){
	gint i;
	boardEmptyPosition(board);
	for(i = a2; i <=h2; i++) boardAddPiece(board, WHITE_PAWN, i);
	boardAddPiece(board, WHITE_ROOK, a1);
	boardAddPiece(board, WHITE_ROOK, h1);
	boardAddPiece(board, WHITE_KNIGHT, b1);
	boardAddPiece(board, WHITE_KNIGHT, g1);
	boardAddPiece(board, WHITE_BISHOP, c1);
	boardAddPiece(board, WHITE_BISHOP, f1);
	boardAddPiece(board, WHITE_QUEEN, d1);
	boardAddPiece(board, WHITE_KING, e1);
	for(i = a7; i <=h7; i++) boardAddPiece(board, BLACK_PAWN, i);
	boardAddPiece(board, BLACK_ROOK, a8);
	boardAddPiece(board, BLACK_ROOK, h8);
	boardAddPiece(board, BLACK_KNIGHT, b8);
	boardAddPiece(board, BLACK_KNIGHT, g8);
	boardAddPiece(board, BLACK_BISHOP, c8);
	boardAddPiece(board, BLACK_BISHOP, f8);
	boardAddPiece(board, BLACK_QUEEN, d8);
	boardAddPiece(board, BLACK_KING, e8);
	board->player = WHITE;
	board->castling = 15;
}

void boardPrint(const Board* board){
	gint piece, index;
	gchar string[72] = {
		'#', ' ', '#', ' ', '#', ' ', '#', ' ', '\n',
		' ', '#', ' ', '#', ' ', '#', ' ', '#', '\n',
		'#', ' ', '#', ' ', '#', ' ', '#', ' ', '\n',
		' ', '#', ' ', '#', ' ', '#', ' ', '#', '\n',
		'#', ' ', '#', ' ', '#', ' ', '#', ' ', '\n',
		' ', '#', ' ', '#', ' ', '#', ' ', '#', '\n',
		'#', ' ', '#', ' ', '#', ' ', '#', ' ', '\n',
		' ', '#', ' ', '#', ' ', '#', ' ', '#', '\0'
	};
	for(index = 0; index<64; index++){
		piece = board->types[index];
		if(piece >= PAWN) string[70-rank(index)*9-(7-file(index))] = pieceToChar[piece];
	}
	g_printf("%s\n\n",string);
	if(board->player == WHITE) g_printf("Player: WHITE\n");
	else g_printf("Player: BLACK\n");
	g_printf("Ply:%d\n",board->ply);
	g_printf("En passant:%d\n",board->enPassant);
	g_printf("Castling:%d\n",board->castling);
	g_printf("Check:%d\n",board->check);
	g_printf("Fifty moves role:%d\n",board50MovesRule(board));
	g_printf("Repeat:%d\n",boardThreeTimeRepeat(board));
	g_printf("Material:%d\n",board->material[WHITE]-board->material[BLACK]);
	g_printf("Key:%u%u\n",(gint32)board->key,(gint32)(board->key>>32));
	g_printf("History: ");
	for(index=0; index<board->ply; index++){
		moveToString(board->historyOfMoves[index],string);
		g_printf("%s ",string);
	}
	g_printf("\n");
}

guint boardThreeTimeRepeat(const Board* board){
	guint i, rp = 0;
	for(i=4; i<=board->numberOfStaticMoves[board->ply]; i+=2){
		if(board->key == board->historyOfKeys[board->ply-i]){
			rp ++;
			i+=2;
		}
	}
	return rp;
}

guint board50MovesRule(const Board* board){
	return board->numberOfStaticMoves[board->ply];
}

/**
 * inline gboolean isInCheck(Board* board)
 * ???
 * \param board ???
 * \return ???
 */
inline gboolean isInCheck(const Board* board){
	#ifndef G_DISABLE_ASSERT
	if(board->king[board->player] == 0){
		boardPrint(board);
		g_error("Bitboard without king!");
	}
	#endif
	guint square = indexOfFirstBit64(&board->king[board->player]);
	if(attackRook(board,square)&(board->rooks[board->opponent]|board->queens[board->opponent]))
	       return TRUE;
	if(attackBishop(board,square)&(board->bishops[board->opponent]|board->queens[board->opponent]))
		return TRUE;
	if(attackKnight[square]&board->knights[board->opponent]) return TRUE;
	if(attackPawn[board->player][square]&board->pawns[board->opponent]) return TRUE;
	return FALSE;
}

inline gboolean isIllegal(const Board* board){
	#ifndef G_DISABLE_ASSERT
	if(board->king[board->player] == 0){
		boardPrint(board);
		g_error("Bitboard without king!");
	}
	#endif
	gint square = indexOfFirstBit64(&board->king[board->opponent]);
	if(attackRook(board,square)&(board->rooks[board->player]|board->queens[board->player]))
		return TRUE;
	if(attackBishop(board,square)&(board->bishops[board->player]|board->queens[board->player]))
		return TRUE;
	if(attackKnight[square]&board->knights[board->player]) return TRUE;
	if(attackPawn[board->opponent][square]&board->pawns[board->player]) return TRUE;
	if(attackKing[square]&board->king[board->player]) return TRUE;
	return FALSE;
}

gboolean boardIsInCheck(const Board* board){
	return board->check;
}

guint boardCurrentPlayer(const Board* board){
	return board->player;
}

void boardSetFen(Board* board, gchar fen[]){
	guint index = 0, line = 7, square = 56, intc;
	boardEmptyPosition(board);
	while(fen[index] == ' ' || fen[index] == '\t') index ++;
	while(fen[index] != ' '){
		intc = (gint)fen[index];
		if(intc >= 48 && intc <=59){ square += intc - 48;}
		else if(fen[index] != '/'){ boardAddPiece(board,pieceFromChar(fen[index]),square++); }
		else square = --line * 8;
		index ++;
	}
	while(fen[index] == ' ' || fen[index] == '\t') index ++;
	if(fen[index] == 'b') changePlayer(board);
	index ++;
	while(fen[index] == ' ' || fen[index] == '\t') index ++;
	while(fen[index] != ' '){
		switch (fen[index]){
			case 'K':{ board->castling |= WHITE_CASTLING_KING_SIDE; break; }
			case 'Q':{ board->castling |= WHITE_CASTLING_QUEEN_SIDE; break; }
			case 'k':{ board->castling |= BLACK_CASTLING_KING_SIDE; break; }
			case 'q':{ board->castling |= BLACK_CASTLING_QUEEN_SIDE; break; }
		}
		index++;
	}
	index ++;
	while(fen[index] == ' ' || fen[index] == '\t') index ++;
	if(fen[index] != '-') board->enPassant = squareFromString(fen+index);
	board->check = isInCheck(board);
	/* Manjkata dva podatka */
}

void boardValidate(const Board* board){
	gint i,type,j;
	gint material=0;
	for(i=WHITE_PAWN; i<=WHITE_QUEEN; i+=2){
		material += numberOfBits64(board->pieces[i])*pieceValue[i];
		material -= numberOfBits64(board->pieces[i+1])*pieceValue[i+1];
	}
	if(material != (board->material[WHITE] - board->material[BLACK])){
		boardPrint(board);
		g_error("Wrong material value (%d) !!!\n",material);
	}
	for(i=0; i<64; i++){
		type = board->types[i];
		if(type == EMPTY) continue;
		if((board->pieces[type] & (1ull<<i)) == 0){
			boardPrint(board);
			g_error("Wrong pieces type (%d) !!!\n",type);
		}
		for(j=WHITE_PAWN; j<type; j++){
			if((board->pieces[j] & (1ull<<i)) != 0){
				boardPrint(board);
				g_error("Doubled bitboard pieces (%d,%d => %d) !!!\n",type,j,i);
			}
		}
		for(j=type+1; j<=BLACK_KING; j++){
			if((board->pieces[j] & (1ull<<i)) != 0){
				boardPrint(board);
				g_error("Doubled bitboard pieces (%d,%d => %d) !!!\n",type,j,i);
			}
		}
	}
	if(board->king[WHITE] == 0){
		boardPrint(board);
		g_error("Board without white king !!!\n");
	}
	if(board->king[BLACK] == 0){
		boardPrint(board);
		g_error("Board without black king !!!\n");
	}
}

#include "boardInit.c"
#include "boardMakeMove.c"
#include "boardMoveGenerator.c"
#include "boardEvaluator.c"
