/***************************************************************************
 *   Copyright (C) 2006 by Borko Boskovic                                  *
 *   borko.boskovic@uni-mb.si                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "boardEndGameEvaluator.c"

gint boardEvaluate(const Board* board, gint alpha,gint beta){
	gint numberOfPieces = board->numberOfPieces[WHITE] + board->numberOfPieces[BLACK];
	if(numberOfPieces <= 4)	return boardEvaluateEndGame(board);
	
	gint result = board->material[WHITE] - board->material[BLACK];
	gint mobility = 0;
	gint from;
	gint dist;
	gint wkf, bkf, wkr, bkr, pf, pr;
	guint64 p;

	if(board->player == WHITE){
		if(result - 300 >= beta) return beta;
		if(result + 300 <= alpha) return alpha;
	}else{
		if(-result - 300 >= beta) return beta;
		if(-result + 300 <= alpha) return alpha;
	}
		
	wkf = file(indexOfFirstBit64(&board->king[WHITE]));
	bkf = file(indexOfFirstBit64(&board->king[BLACK]));
	wkr = rank(indexOfFirstBit64(&board->king[WHITE]));
	bkr = rank(indexOfFirstBit64(&board->king[BLACK]));
		
	if((board->castling & 3) == 0){
		if(board->queens[BLACK] != 0 && board->material[BLACK]>2000){
			if(wkf > 5 && (board->rooks[WHITE] & 192) == 0) result += 50;
			else if (wkf < 3 && (board->rooks[WHITE] & 7) == 0) result += 30;
			else result -= 50;
		}

		p = board->queens[WHITE];
		while(p != 0){
			from = getFirstBitAndClear64(&p);
			mobility += mobilityQueen(board, from);
			pf = file(from);
			pr = rank(from);
			result +=  10 - ((ABS(bkf-pf) + ABS(bkr-pr))); 
			if(pf == 6) result += 10;
		}
		p = board->rooks[WHITE];
		while(p != 0){
			from = getFirstBitAndClear64(&p);
		 	mobility += mobilityRook(board, from);
			pf = file(from);
			if((bitFile[pf] & board->pawns[WHITE]) == 0){
				result += 10;
				if((bitFile[pf] & board->pawns[BLACK]) == 0){
					result += 10;
				}
				dist = ABS(bkf-file(from));
				if(dist <= 1){
					result += 10;
					if(dist == 0) result += 10; 
				}
				if(rank(from) == 6) result += 10;
			}
		}
	}

	if((board->castling & 12) == 0 && numberOfPieces > 13){
		if(board->queens[WHITE] != 0 && board->material[WHITE]>2000){
			if(bkf > 5 && (board->rooks[WHITE] & 13835058055282163712ull) == 0) 
				result -= 50;
			else if(bkf < 3 && (board->rooks[WHITE] & 216172782113783808ull) == 0)
				result -= 30;
			else result += 50;
		}
		
		p = board->queens[BLACK];
		while(p != 0){
			from = getFirstBitAndClear64(&p);
			mobility -= mobilityQueen(board, from);
			pf = file(from);
			pr = rank(from);
			result -=  10 - ((ABS(wkf-pf) + ABS(wkr-pr))); 
			if(pf == 1) result -= 10;
		}
		p = board->rooks[BLACK];
		while(p != 0){
			from = getFirstBitAndClear64(&p);
			mobility -= mobilityRook(board,from);
			pf = file(from);
			if((bitFile[pf] & board->pawns[BLACK]) == 0){
				result -= 10;
				if((bitFile[pf] & board->pawns[WHITE]) == 0){
					result -= 10;
				}
				dist = ABS(wkf-file(from));
				if(dist <= 1){
					result -= 10;
					if(dist == 0) result -= 10; 
				}
				if(rank(from) == 1) result -= 10;
			}
		}
	}
	
	p = board->bishops[WHITE];
	while(p != 0) mobility += mobilityBishop(board, getFirstBitAndClear64(&p));
	
	p = board->bishops[BLACK];
	while(p != 0) mobility -= mobilityBishop(board, getFirstBitAndClear64(&p));
	
	p = board->knights[WHITE];
	while(p != 0) mobility +=mobilityKnight(getFirstBitAndClear64(&p));
		
	p = board->knights[BLACK];
	while(p != 0) mobility -= mobilityKnight(getFirstBitAndClear64(&p));
	
	p = board->pawns[WHITE];
	mobility += numberOfBits64(shiftUp(p) & board->freeSquares);
	result -= numberOfBits64((p & (p << 8)) | (p & (p << 16)) | (p & (p << 24)))<<6;
	result += numberOfBits64(p & RANK5)<<2;
	result += numberOfBits64(p & RANK6)<<4;
	result += numberOfBits64(p & RANK7)<<6;

	p = board->pawns[BLACK];
	mobility -= numberOfBits64(shiftDown(p) & board->freeSquares);
	result += numberOfBits64((p & (p >> 8)) | (p & (p >> 16)) | (p & (p >> 24)))<<6;
	result -= numberOfBits64(p & RANK4)<<2;
	result -= numberOfBits64(p & RANK3)<<4;
	result -= numberOfBits64(p & RANK2)<<6;
	
	result += (mobility<<2);
	
	if(board->player == WHITE) return result;
	else return -result;
}
